# SPDX-License-Identifier: Apache-2.0

from cliff.command import Command

import tempfile
import os
from loguru import logger
import openstack
from tabulate import tabulate
import json
import yaml
from openstack.baremetal import configdrive as configdrive_builder

from osism.commands import get_cloud_connection
from osism import utils


class BaremetalList(Command):
    def get_parser(self, prog_name):
        parser = super(BaremetalList, self).get_parser(prog_name)
        parser.add_argument(
            "--provision-state",
            default=None,
            choices=["enroll", "managable", "available", "active", "error"],
            type=str,
            help="Only list nodes with the given provision state",
        )
        parser.add_argument(
            "--maintenance",
            default=False,
            action="store_true",
            help="Only list baremetal nodes in maintenance mode",
        )
        return parser

    def take_action(self, parsed_args):
        provision_state = parsed_args.provision_state
        maintenance = parsed_args.maintenance

        conn = get_cloud_connection()

        query = {}
        if provision_state:
            query.update(dict(provision_state=provision_state))
        if maintenance:
            query.update(dict(maintenance=maintenance))

        baremetal = conn.baremetal.nodes(**query)

        result = [
            [
                b["name"],
                b["power_state"],
                b["provision_state"],
                b["maintenance"],
            ]
            for b in baremetal
        ]

        print(
            tabulate(
                result,
                headers=[
                    "Name",
                    "Power State",
                    "Provision State",
                    "Maintenance",
                ],
                tablefmt="psql",
            )
        )


class BaremetalDeploy(Command):
    def get_parser(self, prog_name):
        parser = super(BaremetalDeploy, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            nargs="?",
            type=str,
            help="Deploy given baremetal node when in provision state available",
        )
        parser.add_argument(
            "--all",
            default=False,
            help="Deploy all baremetal nodes in provision state available",
            action="store_true",
        )
        parser.add_argument(
            "--rebuild",
            default=False,
            help="Rebuild given nodes in active state",
            action="store_true",
        )
        parser.add_argument(
            "--yes-i-really-really-mean-it",
            default=False,
            help="Specify this in connection with '--rebuild --all' to actually rebuild all nodes",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        all_nodes = parsed_args.all
        name = parsed_args.name
        rebuild = parsed_args.rebuild
        yes_i_really_really_mean_it = parsed_args.yes_i_really_really_mean_it

        if not all_nodes and not name:
            logger.error("Please specify a node name or use --all")
            return

        if all_nodes and rebuild and not yes_i_really_really_mean_it:
            logger.error(
                "Please confirm that you wish to rebuild all nodes by specifying '--yes-i-really-really-mean-it'"
            )
            return

        conn = get_cloud_connection()

        if all_nodes:
            deploy_nodes = list(conn.baremetal.nodes(details=True))
        else:
            node = conn.baremetal.find_node(name, ignore_missing=True, details=True)
            if not node:
                logger.warning(f"Could not find node {name}")
                return
            deploy_nodes = [node]

        for node in deploy_nodes:
            if not node:
                continue

            if (
                node.provision_state in ["available", "deploy failed"]
                and not node["maintenance"]
            ):
                provision_state = "active"
            elif (
                node.provision_state == "error"
                or node.provision_state == "active"
                and not node["maintenance"]
                and rebuild
            ):
                provision_state = "rebuild"
            else:
                logger.warning(
                    f"Node {node.name} ({node.id}) not in supported state! Provision state: {node.provision_state}, maintenance mode: {node['maintenance']}"
                )
                continue

            # NOTE: Ironic removes "instance_info" on undeploy. It was saved to "extra" during sync and needs to be refreshed here.
            if (
                "instance_info" in node
                and not node["instance_info"]
                and "instance_info" in node["extra"]
                and node["extra"]["instance_info"]
            ):
                node = conn.baremetal.update_node(
                    node, instance_info=json.loads(node.extra["instance_info"])
                )

            try:
                conn.baremetal.validate_node(
                    node.id, required=("boot", "deploy", "power")
                )
            except openstack.exceptions.ValidationException:
                logger.warning(f"Node {node.name} ({node.id}) could not be validated")
                continue
            # NOTE: Prepare osism config drive
            try:
                # Get default vars from NetBox local_context_data if available
                default_vars = {}
                if utils.nb:
                    try:
                        # Try to find device by name first
                        device = utils.nb.dcim.devices.get(name=node.name)

                        # If not found by name, try by inventory_hostname custom field
                        if not device:
                            devices = utils.nb.dcim.devices.filter(
                                cf_inventory_hostname=node.name
                            )
                            if devices:
                                device = devices[0]

                        # Extract local_context_data if device found and has the field
                        if (
                            device
                            and hasattr(device, "local_context_data")
                            and device.local_context_data
                        ):
                            default_vars = device.local_context_data
                            logger.info(
                                f"Using NetBox local_context_data for node {node.name}"
                            )
                        else:
                            logger.debug(
                                f"No local_context_data found for node {node.name} in NetBox"
                            )
                    except Exception as e:
                        logger.warning(
                            f"Failed to fetch NetBox data for node {node.name}: {e}"
                        )

                playbook = []
                play = {
                    "name": "Run bootstrap - part 2",
                    "hosts": "localhost",
                    "connection": "local",
                    "gather_facts": True,
                    "vars": default_vars.copy(),
                    "roles": [
                        "osism.commons.hostname",
                        "osism.commons.hosts",
                        "osism.commons.operator",
                    ],
                }
                play["vars"].update(
                    {"hostname_name": node.name, "hosts_type": "template"}
                )
                if (
                    "netplan_parameters" in node.extra
                    and node.extra["netplan_parameters"]
                ):
                    play["vars"].update(
                        {
                            "network_allow_service_restart": True,
                        }
                    )
                    play["vars"].update(json.loads(node.extra["netplan_parameters"]))
                    play["roles"].append("osism.commons.network")
                if "frr_parameters" in node.extra and node.extra["frr_parameters"]:
                    play["vars"].update(
                        {
                            "frr_dummy_interface": "loopback0",
                        }
                    )
                    play["vars"].update(json.loads(node.extra["frr_parameters"]))
                    play["roles"].append("osism.services.frr")
                playbook.append(play)
                with tempfile.TemporaryDirectory() as tmp_dir:
                    with open(os.path.join(tmp_dir, "playbook.yml"), "w") as file:
                        yaml.dump(
                            playbook,
                            file,
                            default_flow_style=False,
                            explicit_start=True,
                            indent=2,
                            sort_keys=False,
                        )
                    config_drive = configdrive_builder.pack(tmp_dir)
            except Exception as exc:
                logger.warning(
                    f"Failed to build config drive for {node.name} ({node.id}): {exc}"
                )
                continue
            try:
                conn.baremetal.set_node_provision_state(
                    node.id, provision_state, config_drive=config_drive
                )
            except Exception as exc:
                logger.warning(
                    f"Node {node.name} ({node.id}) could not be moved to active state: {exc}"
                )
                continue


class BaremetalUndeploy(Command):
    def get_parser(self, prog_name):
        parser = super(BaremetalUndeploy, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            nargs="?",
            type=str,
            help="Undeploy given baremetal node",
        )
        parser.add_argument(
            "--all",
            default=False,
            help="Undeploy all baremetal nodes",
            action="store_true",
        )
        parser.add_argument(
            "--yes-i-really-really-mean-it",
            default=False,
            help="Specify this to actually undeploy all nodes",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        all_nodes = parsed_args.all
        name = parsed_args.name
        yes_i_really_really_mean_it = parsed_args.yes_i_really_really_mean_it

        if not all_nodes and not name:
            logger.error("Please specify a node name or use --all")
            return

        if all_nodes and not yes_i_really_really_mean_it:
            logger.error(
                "Please confirm that you wish to undeploy all nodes by specifying '--yes-i-really-really-mean-it'"
            )
            return

        conn = get_cloud_connection()

        if all_nodes:
            deploy_nodes = list(conn.baremetal.nodes())
        else:
            node = conn.baremetal.find_node(name, ignore_missing=True, details=False)
            if not node:
                logger.warning(f"Could not find node {name}")
                return
            deploy_nodes = [node]

        for node in deploy_nodes:
            if not node:
                continue

            if node.provision_state in ["active", "deploy failed", "error"]:
                try:
                    node = conn.baremetal.set_node_provision_state(node.id, "undeploy")
                except Exception as exc:
                    logger.warning(
                        f"Node {node.name} ({node.id}) could not be moved to available state: {exc}"
                    )
                    continue
            else:
                logger.warning(
                    f"Node {node.name} ({node.id}) not in supported provision state"
                )



# Copyright Notice:
# Copyright 2016-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Mockup-Creator/blob/main/LICENSE.md
# source: https://github.com/DMTF/Redfish-Mockup-Creator

"""
Redfish Mockup Creator

File : redfishMockupCreate.py

Brief : This tool walks a service and creates a mockup from all resources
"""

import argparse
import datetime
import json
import os
import redfish
import sys
import time
import xml.etree.ElementTree as ET
import logging
import copy
import gc
from redfish import redfish_logger

# Version info
tool_version = "1.2.0"

# For Windows, there are restricted characters in folder names that could be used in URIs
disallowed_folder_characters_win = [":", "*", "?", '"', "<", ">", "|"]
folder_name_fix = False
if sys.platform == "win32" or sys.platform == "cygwin":
    folder_name_fix = True


def main():
    """
    Main entry point for the script
    """

    # Get the input arguments
    argget = argparse.ArgumentParser(
        description="A tool to walk a Redfish service and create a mockup from all resources"
    )
    argget.add_argument(
        "--user", "-u", type=str, required=True, help="The user name for authentication"
    )
    argget.add_argument(
        "--password",
        "-p",
        type=str,
        required=True,
        help="The password for authentication",
    )
    argget.add_argument(
        "--rhost",
        "-r",
        type=str,
        required=True,
        help="The IP address (and port) of the Redfish service",
    )
    argget.add_argument(
        "--Dir",
        "-D",
        type=str,
        help="Output directory for the mockup; defaults to 'rfMockUpDfltDir'",
        default="rfMockUpDfltDir",
    )
    argget.add_argument(
        "--Secure", "-S", action="store_true", help="Use HTTPS for all operations"
    )
    argget.add_argument(
        "--Auth",
        "-A",
        type=str,
        help="Authentication mode",
        choices=["None", "Basic", "Session"],
        default="Session",
    )
    argget.add_argument(
        "--Headers",
        "-H",
        action="store_true",
        help="Captures the response headers in the mockup",
    )
    argget.add_argument(
        "--Time",
        "-T",
        action="store_true",
        help="Capture the time of each GET in the mockup",
    )
    argget.add_argument(
        "--Copyright",
        "-C",
        type=str,
        help="Copyright string to add to each resource",
        default=None,
    )
    argget.add_argument(
        "--description",
        "-d",
        type=str,
        help="Mockup description to add to the output readme file",
        default="",
    )
    argget.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Quiet mode; progress messages suppressed",
    )
    argget.add_argument(
        "--trace",
        "-trace",
        action="store_true",
        help="Enable tracing; creates the file rf-mockup-create.log in the output directory to capture Redfish traces with the service",
    )
    argget.add_argument(
        "--maxlogentries",
        "-maxlogentries",
        type=int,
        help="The maximum number of log entries to collect in each log service",
    )
    argget.add_argument(
        "--forcefolderrename",
        "-forcefolderrename",
        action="store_true",
        help="Indicates if URIs containing characters that are disallowed in Windows folder names are renamed to replace the characters with underscores",
    )
    args, unknown = argget.parse_known_args()

    # Convert the authentication method to something usable with the Redfish library
    # This is needed for backwards compatibility with older versions of the tool
    if args.Auth == "Session":
        args.Auth = "session"
    else:
        args.Auth = "basic"

    # Build the base URL for the service
    # More backwards compatibility
    if "://" not in args.rhost:
        if args.Secure:
            args.rhost = "https://{}".format(args.rhost)
        else:
            args.rhost = "http://{}".format(args.rhost)

    # Set up the output
    if not os.path.isdir(args.Dir):
        # Does not exist; make the directory
        try:
            os.makedirs(args.Dir)
        except Exception as err:
            print(
                "ERROR: Aborting; could not create output directory '{}': {}".format(
                    args.Dir, err
                )
            )
            sys.exit(1)
    else:
        if len(os.listdir(args.Dir)) != 0:
            print("ERROR: Aborting; output directory not empty...")
            sys.exit(1)

    print("Redfish Mockup Creator, Version {}".format(tool_version))
    print("Address: {}".format(args.rhost))
    print("Full Output Path: {}".format(os.path.abspath(args.Dir)))
    print("Description: {}".format(args.description))
    print("Starting mockup creation...")
    if args.quiet:
        print("Quiet mode enabled; please wait...")

    # Create the readme file
    try:
        with open(os.path.join(args.Dir, "README"), "w") as readf:
            readf.write(
                "Redfish service captured by the Redfish Mockup Creator, Version {}\n".format(
                    tool_version
                )
            )
            readf.write(
                "Created: {}\n".format(
                    datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                )
            )
            readf.write("Service: {}\n".format(args.rhost))
            readf.write("User: {}\n".format(args.user))
            readf.write("Description: {}\n".format(args.description))
    except Exception as err:
        print(
            "ERROR: Aborting; could not create README file in output directory: {}".format(
                err
            )
        )
        sys.exit(1)

    # Set up the trace file if requested
    if args.trace:
        redfish_logger(
            os.path.join(args.Dir, "rf-mockup-create.log"),
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            logging.DEBUG,
        )

    # Set up the Redfish object
    try:
        redfish_obj = redfish.redfish_client(
            base_url=args.rhost, username=args.user, password=args.password
        )
        redfish_obj.login(auth=args.Auth)
    except Exception as err:
        print(
            "ERROR: Aborting; could not authenticate with the Redfish service: {}".format(
                err
            )
        )
        sys.exit(1)

    # Scan the service
    response_times = {}
    scan_resource(redfish_obj, args, response_times, "/redfish")
    scan_resource(redfish_obj, args, response_times, "/redfish/v1/odata")
    scan_resource(
        redfish_obj, args, response_times, "/redfish/v1/$metadata", is_csdl=True
    )
    scan_resource(redfish_obj, args, response_times, "/redfish/v1")
    redfish_obj.logout()

    # Add time statistics to the readme
    total_response_time = sum(response_times.values())
    average_response_time = total_response_time / len(response_times)
    min_response_uri = min(response_times, key=response_times.get)
    max_response_uri = max(response_times, key=response_times.get)
    with open(os.path.join(args.Dir, "README"), "a") as readf:
        readf.write("Total response time: {}\n".format(total_response_time))
        readf.write("Average response time: {}\n".format(average_response_time))
        readf.write(
            "Minimum response time: {}, {}\n".format(
                response_times[min_response_uri], min_response_uri
            )
        )
        readf.write(
            "Maximum response time: {}, {}\n".format(
                response_times[max_response_uri], max_response_uri
            )
        )

    print("Completed mockup creation!")


def scan_resource(redfish_obj, args, response_times, uri, is_csdl=False):
    """
    Scans a resource and saves its response

    Args:
        redfish_obj: The Redfish client object with an open session
        args: The command line arguments
        response_times: The response times database
        uri: The URI to get
        is_csdl: Indicates if the resource is a CSDL file
    """

    # Check if the URI is a relative URI
    if not uri.startswith("/"):
        return

    # Set up the output folder
    try:
        path = uri[1:]
        if folder_name_fix or args.forcefolderrename:
            for character in disallowed_folder_characters_win:
                path = path.replace(character, "_")
        path = os.path.join(args.Dir, path)
        if not os.path.isdir(path):
            # Does not exist; make the directory
            os.makedirs(path)
    except Exception as err:
        print("ERROR: Could not create directory for '{}': {}".format(uri, err))
        return

    # Check if the index file already exists
    index_name = "index.json"
    if is_csdl:
        index_name = "index.xml"
    index_path = os.path.join(path, index_name)
    if os.path.isfile(index_path):
        # File exists; already scanned this resource
        return

    # Get the resource
    if not args.quiet:
        print("Getting {}...".format(uri))
    try:
        start_time = time.time()
        resource = redfish_obj.get(uri, headers={"Accept-Encoding": "*"})
        end_time = time.time()
    except Exception as err:
        print("ERROR: Could not get '{}': {}".format(uri, err))
        return

    # Save the resource and other information
    try:
        # Save the resource itself
        if is_csdl:
            with open(index_path, "w", encoding="utf-8") as file:
                file.write(resource.text)
        else:
            save_dict = resource.dict

            # Prune the log entry collection if needed
            if (
                save_dict.get("@odata.type", None)
                == "#LogEntryCollection.LogEntryCollection"
                and args.maxlogentries is not None
            ):
                if args.maxlogentries < 0:
                    args.maxlogentries = 0
                if "Members@odata.nextLink" in save_dict:
                    save_dict.pop("Members@odata.nextLink")
                if "Members" in save_dict:
                    if isinstance(save_dict["Members"], list):
                        for i in range(
                            0, len(save_dict["Members"]) - args.maxlogentries
                        ):
                            save_dict["Members"].pop()
                        save_dict["Members@odata.count"] = len(save_dict["Members"])

            # The saved copy might contain URI fixes and other changes that aren't reflective of the service, but are
            # needed to ensure compatibility with the system creating the mockup
            scan_dict = copy.deepcopy(save_dict)

            # Add the copyright statement if needed
            if args.Copyright:
                save_dict["@Redfish.Copyright"] = args.Copyright

            # Update the payload's URIs if they need to be corrected based on allowable folder names for the system
            if folder_name_fix or args.forcefolderrename:
                fix_uris(save_dict)

            with open(index_path, "w", encoding="utf-8") as file:
                json.dump(save_dict, file, indent=4, separators=(",", ": "))

            # Deep copies of all payloads gets expensive; force garbage collection to avoid stack overflows
            del save_dict
            gc.collect()
    except Exception as err:
        print("ERROR: Could not save '{}': {}".format(uri, err))
        print("Attempting to save response data in error.txt...")
        try:
            with open(os.path.join(path, "error.txt"), "w", encoding="utf-8") as file:
                file.write("HTTP {}\n".format(resource.status))
                for header in resource.getheaders():
                    file.write("{}: {}\n".format(header[0], header[1]))
                file.write("\n")
                file.write(resource.text)
        except:
            print("Could not save response data; moving on... ")
        return

    # Save additional info
    try:
        # Save headers
        if args.Headers:
            with open(
                os.path.join(path, "headers.json"), "w", encoding="utf-8"
            ) as file:
                headers_dict = {}
                for header in resource.getheaders():
                    headers_dict[header[0]] = header[1]
                json.dump({"GET": headers_dict}, file, indent=4, separators=(",", ": "))

        # Save timing info
        response_times[uri] = end_time - start_time
        if args.Time:
            with open(os.path.join(path, "time.json"), "w", encoding="utf-8") as file:
                json.dump(
                    {"GET_Time": "{0:.2f}".format(response_times[uri])},
                    file,
                    indent=4,
                    separators=(",", ": "),
                )
    except Exception as err:
        print(
            "ERROR: Could not save header or timing data for '{}': {}".format(uri, err)
        )
        return

    # Scan the response to see where to go next
    try:
        if is_csdl:
            scan_csdl(redfish_obj, args, response_times, resource.text)
        else:
            scan_object(redfish_obj, args, response_times, scan_dict)
    except Exception as err:
        print("ERROR: Could not scan '{}': {}".format(uri, err))
        return


def scan_object(redfish_obj, args, response_times, object):
    """
    Scans an object or array to find links to other resources

    Args:
        redfish_obj: The Redfish client object with an open session
        args: The command line arguments
        response_times: The response times database
        object: The object to scan
    """

    for item in object:
        # If the object is a dictionary, inspect the properties found
        if isinstance(object, dict):
            # If the item is a reference, go to the resource
            if (
                item == "@odata.id"
                or item == "Uri"
                or item == "Members@odata.nextLink"
                or item == "@Redfish.ActionInfo"
            ):
                if isinstance(object[item], str):
                    if object[item].startswith("/") and "#" not in object[item]:
                        scan_resource(redfish_obj, args, response_times, object[item])

            # If the item is an object or array, scan one level deeper
            elif isinstance(object[item], dict) or isinstance(object[item], list):
                scan_object(redfish_obj, args, response_times, object[item])

        # If the object is a list, see if the member needs to be scanned
        elif isinstance(object, list):
            if isinstance(item, dict) or isinstance(item, list):
                scan_object(redfish_obj, args, response_times, item)


def scan_csdl(redfish_obj, args, response_times, csdl):
    """
    Scans a CSDL string to find links to other CSDL files

    Args:
        redfish_obj: The Redfish client object with an open session
        args: The command line arguments
        response_times: The response times database
        csdl: The CSDL string to scan
    """

    # Convert to an element tree object
    tree = ET.ElementTree(ET.fromstring(csdl))
    root = tree.getroot()

    # Find references
    for reference in root:
        if "reference" in str(reference.tag).lower():
            if "Reference" not in str(reference.tag):
                print(
                    "Warning: Found invalid reference tag '{}'; tags are case sensitive!".format(
                        str(reference.tag)
                    )
                )
            for tag in ["Uri", "uri", "URI"]:
                uri = reference.attrib.get(tag)
                if uri is not None:
                    if tag != "Uri":
                        print(
                            "Warning: Found invalid Uri attribute '{}'; attributes are case sensitive!".format(
                                tag
                            )
                        )
                    # Scan the reference
                    scan_resource(redfish_obj, args, response_times, uri, is_csdl=True)


def fix_uris(payload):
    """
    Updates URIs in a payload to ensure they do not conflict with local system folder name rules

    Args:
        payload: The payload to update
    """

    for item in payload:
        # If the payload is a dictionary, inspect the properties found
        if isinstance(payload, dict):
            # If the item is a reference, go to the resource
            if item == "@odata.id" or item == "Uri" or item == "Members@odata.nextLink":
                if isinstance(payload[item], str):
                    for character in disallowed_folder_characters_win:
                        payload[item] = payload[item].replace(character, "_")

            # If the item is an object or array, scan one level deeper
            elif isinstance(payload[item], dict) or isinstance(payload[item], list):
                fix_uris(payload[item])

        # If the object is a list, see if the member needs to be scanned
        elif isinstance(payload, list):
            if isinstance(item, dict) or isinstance(item, list):
                fix_uris(item)


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: Apache-2.0

"""Interface caching for SONiC configuration generation."""

import threading
from typing import Dict, List, Optional
from loguru import logger

from osism import utils


class InterfaceCache:
    """Thread-local cache for device interfaces during sync_sonic task."""

    def __init__(self):
        self._cache: Dict[int, List] = {}
        self._lock = threading.Lock()

    def get_device_interfaces(self, device_id: int) -> List:
        """Get interfaces for a device, using cache if available.

        Args:
            device_id: NetBox device ID

        Returns:
            List of interface objects
        """
        with self._lock:
            if device_id not in self._cache:
                logger.debug(f"Fetching interfaces for device {device_id}")
                try:
                    interfaces = list(
                        utils.nb.dcim.interfaces.filter(device_id=device_id)
                    )
                    self._cache[device_id] = interfaces
                    logger.debug(
                        f"Cached {len(interfaces)} interfaces for device {device_id}"
                    )
                except Exception as e:
                    logger.warning(
                        f"Failed to fetch interfaces for device {device_id}: {e}"
                    )
                    self._cache[device_id] = []
            else:
                logger.debug(f"Using cached interfaces for device {device_id}")

            return self._cache[device_id]

    def clear(self):
        """Clear the cache."""
        with self._lock:
            cache_size = len(self._cache)
            self._cache.clear()
            logger.debug(f"Cleared interface cache ({cache_size} devices)")

    def get_cache_stats(self) -> Dict[str, int]:
        """Get cache statistics.

        Returns:
            Dictionary with cache statistics
        """
        with self._lock:
            total_interfaces = sum(
                len(interfaces) for interfaces in self._cache.values()
            )
            return {
                "cached_devices": len(self._cache),
                "total_interfaces": total_interfaces,
            }


# Thread-local storage for the interface cache
_thread_local = threading.local()


def get_interface_cache() -> InterfaceCache:
    """Get the current thread's interface cache.

    Returns:
        InterfaceCache instance for current thread
    """
    if not hasattr(_thread_local, "interface_cache"):
        _thread_local.interface_cache = InterfaceCache()
    return _thread_local.interface_cache


def get_cached_device_interfaces(device_id: int) -> List:
    """Get interfaces for a device using the thread-local cache.

    Args:
        device_id: NetBox device ID

    Returns:
        List of interface objects
    """
    cache = get_interface_cache()
    return cache.get_device_interfaces(device_id)


def clear_interface_cache():
    """Clear the current thread's interface cache."""
    if hasattr(_thread_local, "interface_cache"):
        _thread_local.interface_cache.clear()


def get_interface_cache_stats() -> Optional[Dict[str, int]]:
    """Get cache statistics for the current thread.

    Returns:
        Dictionary with cache statistics or None if no cache exists
    """
    if hasattr(_thread_local, "interface_cache"):
        return _thread_local.interface_cache.get_cache_stats()
    return None



# SPDX-License-Identifier: Apache-2.0

from cliff.command import Command
from loguru import logger

from osism.tasks import ansible, conductor, handle_task


class Facts(Command):
    def get_parser(self, prog_name):
        parser = super(Facts, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        arguments = []
        t = ansible.run.delay(
            "generic", "gather-facts", arguments, auto_release_time=3600
        )
        rc = handle_task(t)
        return rc


class Sonic(Command):
    def get_parser(self, prog_name):
        parser = super(Sonic, self).get_parser(prog_name)
        parser.add_argument(
            "device",
            nargs="?",
            help="Optional device name to sync configuration for a specific device",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        parser.add_argument(
            "--diff",
            default=True,
            help="Show configuration diff when changes are detected (default: True)",
            action="store_true",
        )
        parser.add_argument(
            "--no-diff",
            dest="diff",
            help="Do not show configuration diff",
            action="store_false",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait
        device_name = parsed_args.device
        show_diff = parsed_args.diff

        task = conductor.sync_sonic.delay(device_name, show_diff)

        if device_name:
            logger.info(
                f"Task {task.task_id} (sync sonic for device {device_name}) started"
            )
        else:
            logger.info(f"Task {task.task_id} (sync sonic) started")

        rc = handle_task(task, wait=wait)
        return rc



# SPDX-License-Identifier: Apache-2.0

import os
import time
from collections.abc import Callable
from typing import Any

from kombu import Connection, Exchange, Queue
from kombu.mixins import ConsumerMixin
from loguru import logger
import json
import requests

from osism.tasks import netbox
from osism import settings

EXCHANGE_NAME = "ironic"
ROUTING_KEY = "ironic_versioned_notifications.info"
QUEUE_NAME = "osism-listener-ironic"
BROKER_URI = os.getenv("BROKER_URI")


class BaremetalEvents:
    # References:
    #
    # * https://docs.openstack.org/ironic/latest/admin/notifications.html#available-notifications
    # * https://docs.openstack.org/ironic/latest/user/states.html
    # * https://docs.openstack.org/ironic/latest/_images/states.svg

    def __init__(self) -> None:
        self._handler: dict[
            str, dict[str, dict[str, dict[str, Callable[[dict[Any, Any]], None]]]]
        ] = {
            "baremetal": {
                "node": {
                    "power_set": {"end": self.node_power_set_end},
                    "power_state_corrected": {
                        "success": self.node_power_state_corrected_success
                    },
                    "maintenance_set": {"end": self.node_maintenance_set_end},
                    "provision_set": {
                        "start": self.node_provision_set_start,
                        "end": self.node_provision_set_end,
                        "success": self.node_provision_set_success,
                    },
                    "delete": {"end": self.node_delete_end},
                    "create": {"end": self.node_create_end},
                },
            }
        }

    def get_object_data(self, payload: dict[Any, Any]) -> Any:
        return payload["ironic_object.data"]

    def get_handler(self, event_type: str) -> Callable[[dict[Any, Any]], None]:
        event_type_keys = event_type.split(".")
        try:
            handler = self._handler[event_type_keys[0]][event_type_keys[1]][
                event_type_keys[2]
            ][event_type_keys[3]]
        except KeyError:

            def default_handler(payload: dict[Any, Any]) -> None:
                object_data = self.get_object_data(payload)
                name = object_data["name"]
                logger.info(event_type + f" ## {name}")

            handler = default_handler
        return handler

    def node_power_set_end(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.power_set.end ## {name} ## {object_data['power_state']}"
        )
        netbox.set_power_state.delay(name, object_data["power_state"])

    def node_power_state_corrected_success(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.power_state_corrected.success ## {name} ## {object_data['power_state']}"
        )
        netbox.set_power_state.delay(name, object_data["power_state"])

    def node_maintenance_set_end(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.maintenance_set.end ## {name} ## {object_data['maintenance']}"
        )
        netbox.set_maintenance.delay(name, state=object_data["maintenance"])

    def node_provision_set_success(self, payload: dict[Any, Any]) -> None:
        # A provision status was successfully set, update it in the NetBox
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.provision_set.success ## {name} ## {object_data['provision_state']}"
        )
        netbox.set_provision_state.delay(name, object_data["provision_state"])

    def node_provision_set_start(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.provision_set.start ## {name} ## {object_data['provision_state']}"
        )
        netbox.set_provision_state.delay(name, object_data["provision_state"])

    def node_provision_set_end(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(
            f"baremetal.node.provision_set.end ## {name} ## {object_data['provision_state']}"
        )
        netbox.set_provision_state.delay(name, object_data["provision_state"])

    def node_delete_end(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(f"baremetal.node.delete.end ## {name}")
        netbox.set_provision_state.delay(name, None)
        netbox.set_power_state.delay(name, None)

    def node_create_end(self, payload: dict[Any, Any]) -> None:
        object_data = self.get_object_data(payload)
        name = object_data["name"]
        logger.info(f"baremetal.node.create.end ## {name}")
        netbox.set_provision_state.delay(name, object_data["provision_state"])
        netbox.set_power_state.delay(name, object_data["power_state"])


class NotificationsDump(ConsumerMixin):
    def __init__(self, connection):
        self.connection = connection
        self.baremetal_events = BaremetalEvents()
        self.osism_api_session: None | requests.Session = None
        self.osism_baremetal_api_url: None | str = None
        if settings.OSISM_API_URL:
            logger.info("Setting up OSISM API")
            self.osism_api_session = requests.Session()
            self.osism_baremetal_api_url = (
                settings.OSISM_API_URL.rstrip("/") + "/notifications/baremetal"
            )
        return

    def get_consumers(self, consumer, channel):
        exchange = Exchange(EXCHANGE_NAME, type="topic", durable=False)
        queue = Queue(
            QUEUE_NAME,
            exchange,
            routing_key=ROUTING_KEY,
            durable=False,
            auto_delete=True,
            no_ack=True,
        )
        return [consumer(queue, callbacks=[self.on_message])]

    def on_message(self, body, message):
        data = json.loads(body["oslo.message"])
        logger.debug(
            data["event_type"]
            + ": "
            + str(
                {
                    k: v
                    for k, v in data["payload"]["ironic_object.data"].items()
                    if k in ["provision_state", "power_state"]
                }
            )
        )

        if self.osism_api_session:
            tries = 1
            max_tries = 3
            while tries <= max_tries:
                logger.info(
                    f"Trying to deliver notification to {self.osism_baremetal_api_url} (Try: {tries}/{max_tries})\n"
                )
                try:
                    response = self.osism_api_session.post(
                        self.osism_baremetal_api_url,
                        timeout=5,
                        json=dict(
                            priority=data["priority"],
                            event_type=data["event_type"],
                            timestamp=data["timestamp"],
                            publisher_id=data["publisher_id"],
                            message_id=data["message_id"],
                            payload=data["payload"],
                        ),
                    )
                    if response.status_code == 204:
                        logger.info(
                            f"Successfully delivered notification to {self.osism_baremetal_api_url} (Try: {tries}/{max_tries})"
                        )
                        break
                    else:
                        response.raise_for_status()
                except requests.ConnectionError:
                    logger.error(f"Error connecting to {self.osism_baremetal_api_url}")
                except requests.Timeout:
                    logger.error(
                        f"Timeout reached while connecting to {self.osism_baremetal_api_url}"
                    )
                except requests.HTTPError as e:
                    logger.error(
                        f"Received HTTP status code {e.response.status_code} while connecting to {self.osism_baremetal_api_url}"
                    )
                    if e.response.status_code <= 500:
                        logger.error(
                            f"Received HTTP status code {e.response.status_code} indicates a client side error, giving up early"
                        )
                        break

                logger.error(
                    f"Failed to deliver notification to {self.osism_baremetal_api_url} ({tries}/{max_tries})"
                )
                tries += 1
                if tries > max_tries:
                    logger.error(
                        f"Giving up delivering notification to {self.osism_baremetal_api_url} with data:\n"
                        + json.dumps(data)
                    )
                else:
                    time.sleep(pow(3, tries - 1))

        else:
            handler = self.baremetal_events.get_handler(data["event_type"])
            handler(data["payload"])


def main():
    while True:
        try:
            with Connection(BROKER_URI, connect_timeout=30.0) as connection:
                connection.connect()
                NotificationsDump(connection).run()
        except ConnectionRefusedError:
            logger.error("Connection with broker refused. Retry in 60 seconds.")
            time.sleep(60)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0

from celery import Celery
from loguru import logger

from osism import settings, utils
from osism.tasks import Config, run_command

app = Celery("netbox")
app.config_from_object(Config)


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.netbox.run")
def run(self, action, arguments):
    pass


# NOTE: While `get_*` tasks only operate on the NetBox configured in NETBOX_URL,
#       `set_*` tasks additionally operate on all NetBox instances listed in
#       NETBOX_SECONDARIES


@app.task(bind=True, name="osism.tasks.netbox.set_maintenance")
def set_maintenance(self, device_name, state=True):
    """Set the maintenance state for a device in the NetBox."""

    lock = utils.create_redlock(
        key=f"lock_osism_tasks_netbox_set_maintenance_{device_name}",
        auto_release_time=60,
    )
    if lock.acquire(timeout=20):
        try:
            for nb in [utils.nb] + utils.secondary_nb_list:
                logger.info(
                    f"Set maintenance state of device {device_name} = {state} on {nb.base_url}"
                )
                device = nb.dcim.devices.get(name=device_name)
                if device:
                    device.custom_fields.update({"maintenance": state})
                    device.save()
                else:
                    logger.error(
                        f"Could not set maintenance for {device_name} on {nb.base_url}"
                    )
        finally:
            lock.release()
    else:
        logger.error("Could not acquire lock for node {device_name}")


@app.task(bind=True, name="osism.tasks.netbox.set_provision_state")
def set_provision_state(self, device_name, state):
    """Set the provision state for a device in the NetBox."""

    lock = utils.create_redlock(
        key=f"lock_osism_tasks_netbox_set_provision_state_{device_name}",
        auto_release_time=60,
    )
    if lock.acquire(timeout=20):
        try:

            for nb in [utils.nb] + utils.secondary_nb_list:
                logger.info(
                    f"Set provision state of device {device_name} = {state} on {nb.base_url}"
                )
                device = nb.dcim.devices.get(name=device_name)
                if device:
                    device.custom_fields.update({"provision_state": state})
                    device.save()
                else:
                    logger.error(
                        f"Could not set provision state for {device_name} on {nb.base_url}"
                    )
        finally:
            lock.release()
    else:
        logger.error("Could not acquire lock for node {device_name}")


@app.task(bind=True, name="osism.tasks.netbox.set_power_state")
def set_power_state(self, device_name, state):
    """Set the provision state for a device in the NetBox."""

    lock = utils.create_redlock(
        key=f"lock_osism_tasks_netbox_set_provision_state_{device_name}",
        auto_release_time=60,
    )
    if lock.acquire(timeout=20):
        try:
            for nb in [utils.nb] + utils.secondary_nb_list:
                logger.info(
                    f"Set power state of device {device_name} = {state} on {nb.base_url}"
                )
                device = nb.dcim.devices.get(name=device_name)
                if device:
                    device.custom_fields.update({"power_state": state})
                    device.save()
                else:
                    logger.error(
                        f"Could not set power state for {device_name} on {nb.base_url}"
                    )
        finally:
            lock.release()
    else:
        logger.error("Could not acquire lock for node {device_name}")


@app.task(bind=True, name="osism.tasks.netbox.get_location_id")
def get_location_id(self, location_name):
    try:
        location = utils.nb.dcim.locations.get(name=location_name)
    except ValueError:
        return None
    if location:
        return location.id
    else:
        return None


@app.task(bind=True, name="osism.tasks.netbox.get_rack_id")
def get_rack_id(self, rack_name):
    try:
        rack = utils.nb.dcim.racks.get(name=rack_name)
    except ValueError:
        return None
    if rack:
        return rack.id
    else:
        return None


@app.task(bind=True, name="osism.tasks.netbox.get_devices")
def get_devices(self, **query):
    return utils.nb.dcim.devices.filter(**query)


@app.task(bind=True, name="osism.tasks.netbox.get_device_by_name")
def get_device_by_name(self, name):
    return utils.nb.dcim.devices.get(name=name)


@app.task(bind=True, name="osism.tasks.netbox.get_interfaces_by_device")
def get_interfaces_by_device(self, device_name):
    return utils.nb.dcim.interfaces.filter(device=device_name)


@app.task(bind=True, name="osism.tasks.netbox.get_addresses_by_device_and_interface")
def get_addresses_by_device_and_interface(self, device_name, interface_name):
    return utils.nb.dcim.addresses.filter(device=device_name, interface=interface_name)


@app.task(bind=True, name="osism.tasks.netbox.manage")
def manage(self, *arguments, publish=True, locking=False, auto_release_time=3600):
    netbox_manager_env = {
        "NETBOX_MANAGER_URL": str(settings.NETBOX_URL),
        "NETBOX_MANAGER_TOKEN": str(settings.NETBOX_TOKEN),
        "NETBOX_MANAGER_IGNORE_SSL_ERRORS": str(settings.IGNORE_SSL_ERRORS),
        "NETBOX_MANAGER_VERBOSE": "true",
    }

    return run_command(
        self.request.id,
        "/usr/local/bin/netbox-manager",
        netbox_manager_env,
        *arguments,
        publish=publish,
        locking=locking,
        auto_release_time=auto_release_time,
    )


@app.task(bind=True, name="osism.tasks.netbox.ping")
def ping(self):
    status = utils.nb.status()

    return status



# SPDX-License-Identifier: Apache-2.0

"""SONiC configuration management package."""

from .config_generator import generate_sonic_config
from .exporter import save_config_to_netbox, export_config_to_file
from .sync import sync_sonic
from .connections import (
    get_connected_interfaces,
    get_connected_device_for_sonic_interface,
    get_connected_device_via_interface,
    find_interconnected_devices,
    get_device_bgp_neighbors_via_loopback,
)

__all__ = [
    "generate_sonic_config",
    "save_config_to_netbox",
    "export_config_to_file",
    "sync_sonic",
    "get_connected_interfaces",
    "get_connected_device_for_sonic_interface",
    "get_connected_device_via_interface",
    "find_interconnected_devices",
    "get_device_bgp_neighbors_via_loopback",
]



# SPDX-License-Identifier: Apache-2.0

TEMPLATE_IMAGE_OCTAVIA = """---
images:
  - name: OpenStack Octavia Amphora
    enable: true
    shortname: amphora
    format: qcow2
    login: ubuntu
    min_disk: 2
    min_ram: 512
    status: active
    visibility: private
    multi: false
    meta:
      architecture: x86_64
      hw_disk_bus: scsi
      hw_rng_model: virtio
      hw_scsi_model: virtio-scsi
      hw_watchdog_action: reset
      hypervisor_type: qemu
      os_distro: ubuntu
      replace_frequency: quarterly
      uuid_validity: last-1
      provided_until: none
    tags:
      - amphora
    versions:
      - version: "{{ image_version }}"
        url: "{{ image_url }}"
        checksum: "{{ image_checksum }}"
        build_date: {{ image_builddate }}

"""

TEMPLATE_IMAGE_CLUSTERAPI = """---
images:
  - name: ubuntu-capi-image
    enable: true
    keep: true
    separator: "-"
    format: qcow2
    login: ubuntu
    min_disk: 20
    min_ram: 512
    status: active
    visibility: public
    multi: false
    meta:
      architecture: x86_64
      hw_disk_bus: scsi
      hw_rng_model: virtio
      hw_scsi_model: virtio-scsi
      hw_watchdog_action: reset
      hypervisor_type: qemu
      os_distro: ubuntu
      replace_frequency: never
      uuid_validity: none
      provided_until: none
    tags: []
    versions:
      - version: "v{{ image_version }}"
        url: "{{ image_url }}"
        checksum: "{{ image_checksum }}"
        build_date: {{ image_builddate }}

"""



# SPDX-License-Identifier: Apache-2.0

from osism.tasks.conductor import (
    app,
    get_ironic_parameters,
    sync_netbox,
    sync_ironic,
    sync_sonic,
)

__all__ = [
    "app",
    "get_ironic_parameters",
    "sync_netbox",
    "sync_ironic",
    "sync_sonic",
]



# SPDX-License-Identifier: Apache-2.0

import time
import os
from contextlib import redirect_stdout, redirect_stderr
from cryptography.fernet import Fernet
import keystoneauth1
from loguru import logger
import openstack
import pynetbox
from pottery import Redlock
from redis import Redis
import urllib3
import yaml

from osism import settings


def get_netbox_connection(netbox_url, netbox_token, ignore_ssl_errors=False):
    if netbox_url and netbox_token:
        nb = pynetbox.api(netbox_url, token=netbox_token)

        if ignore_ssl_errors and nb:
            import requests

            urllib3.disable_warnings()
            session = requests.Session()
            session.verify = False
            nb.http_session = session

    else:
        nb = None

    return nb


redis = Redis(
    host=settings.REDIS_HOST,
    port=settings.REDIS_PORT,
    db=settings.REDIS_DB,
    socket_keepalive=True,
)
redis.ping()

nb = get_netbox_connection(
    settings.NETBOX_URL, settings.NETBOX_TOKEN, settings.IGNORE_SSL_ERRORS
)

try:
    secondary_nb_settings_list = yaml.safe_load(settings.NETBOX_SECONDARIES)
    supported_secondary_nb_keys = ["NETBOX_URL", "NETBOX_TOKEN", "IGNORE_SSL_ERRORS"]
    secondary_nb_list = []
    if type(secondary_nb_settings_list) is not list:
        raise TypeError(
            f"Setting NETBOX_SECONDARIES needs to be an array of mappings containing supported NetBox API configuration: {supported_secondary_nb_keys}"
        )
    for secondary_nb_settings in secondary_nb_settings_list:
        if type(secondary_nb_settings) is not dict:
            raise TypeError(
                f"Elements in setting NETBOX_SECONDARIES need to be mappings containing supported NetBox API configuration: {supported_secondary_nb_keys}"
            )
        for key in list(secondary_nb_settings.keys()):
            if key not in supported_secondary_nb_keys:
                raise ValueError(
                    f"Unknown key in element of setting NETBOX_SECONDARIES. Supported keys: {supported_secondary_nb_keys}"
                )
        if (
            "NETBOX_URL" not in secondary_nb_settings
            or not secondary_nb_settings["NETBOX_URL"]
        ):
            raise ValueError(
                "All NETBOX_URL values in the elements of setting NETBOX_SECONDARIES need to be valid NetBox URLs"
            )
        if (
            "NETBOX_TOKEN" not in secondary_nb_settings
            or not secondary_nb_settings["NETBOX_TOKEN"]
        ):
            raise ValueError(
                "All NETBOX_TOKEN values in the elements of setting NETBOX_SECONDARIES need to be valid NetBox tokens"
            )

        secondary_nb_list.append(
            get_netbox_connection(
                secondary_nb_settings["NETBOX_URL"],
                secondary_nb_settings["NETBOX_TOKEN"],
                secondary_nb_settings.get("IGNORE_SSL_ERRORS", True),
            )
        )
except (yaml.YAMLError, TypeError, ValueError) as exc:
    logger.error(f"Error parsing settings NETBOX_SECONDARIES: {exc}")
    secondary_nb_list = []


def get_openstack_connection():
    try:
        conn = openstack.connect()
    except keystoneauth1.exceptions.auth_plugins.MissingRequiredOptions:
        pass

    return conn


def get_ansible_vault_password():
    keyfile = "/share/ansible_vault_password.key"

    try:
        with open(keyfile, "r") as fp:
            key = fp.read()
        f = Fernet(key)

        encrypted_ansible_vault_password = redis.get("ansible_vault_password")
        ansible_vault_password = f.decrypt(encrypted_ansible_vault_password)
        return ansible_vault_password.decode("utf-8")
    except Exception as exc:
        logger.error("Unable to get ansible vault password")
        raise exc


# https://stackoverflow.com/questions/2361426/get-the-first-item-from-an-iterable-that-matches-a-condition
def first(iterable, condition=lambda x: True):
    """
    Returns the first item in the `iterable` that
    satisfies the `condition`.

    If the condition is not given, returns the first item of
    the iterable.

    Raises `StopIteration` if no item satysfing the condition is found.

    >>> first( (1,2,3), condition=lambda x: x % 2 == 0)
    2
    >>> first(range(3, 100))
    3
    >>> first( () )
    Traceback (most recent call last):
    ...
    StopIteration
    """

    return next(x for x in iterable if condition(x))


def fetch_task_output(
    task_id, timeout=os.environ.get("OSISM_TASK_TIMEOUT", 300), enable_play_recap=False
):
    rc = 0
    stoptime = time.time() + timeout
    last_id = 0
    while time.time() < stoptime:
        data = redis.xread({str(task_id): last_id}, count=1, block=(timeout * 1000))
        if data:
            stoptime = time.time() + timeout
            messages = data[0]
            for message_id, message in messages[1]:
                last_id = message_id.decode()
                message_type = message[b"type"].decode()
                message_content = message[b"content"].decode()

                logger.debug(f"Processing message {last_id} of type {message_type}")
                redis.xdel(str(task_id), last_id)

                if message_type == "stdout":
                    print(message_content, end="")
                    if enable_play_recap and "PLAY RECAP" in message_content:
                        logger.info(
                            "Play has been completed. There may now be a delay until "
                            "all logs have been written."
                        )
                        logger.info("Please wait and do not abort execution.")
                elif message_type == "rc":
                    rc = int(message_content)
                elif message_type == "action" and message_content == "quit":
                    redis.close()
                    return rc
    raise TimeoutError


def push_task_output(task_id, line):
    redis.xadd(task_id, {"type": "stdout", "content": line})


def finish_task_output(task_id, rc=None):
    if rc:
        redis.xadd(task_id, {"type": "rc", "content": rc})
    redis.xadd(task_id, {"type": "action", "content": "quit"})


def create_redlock(key, auto_release_time=3600):
    """
    Create a Redlock instance with output suppression during initialization.

    Args:
        key (str): The lock key
        auto_release_time (int): Auto release time in seconds (default: 3600)

    Returns:
        Redlock: The configured Redlock instance
    """
    import logging

    # Permanently suppress pottery logger output
    pottery_logger = logging.getLogger("pottery")
    pottery_logger.setLevel(logging.CRITICAL)

    with open(os.devnull, "w") as devnull:
        with redirect_stdout(devnull), redirect_stderr(devnull):
            return Redlock(
                key=key,
                masters={redis},
                auto_release_time=auto_release_time,
            )



# SPDX-License-Identifier: Apache-2.0

import time
import datetime

from cliff.command import Command
from jc import parse
from loguru import logger
import openstack
from tabulate import tabulate
from prompt_toolkit import prompt

from osism.commands import get_cloud_connection, get_cloud_project


class ComputeEnable(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeEnable, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host to be enabled",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        conn = get_cloud_connection()

        services = conn.compute.services(**{"host": host, "binary": "nova-compute"})
        service = next(services)

        if service["forced_down"]:
            logger.info(
                f"Remove force down from nova-compute binary @ {host} ({service.id})"
            )

            try:
                conn.compute.update_service_forced_down(
                    service=service.id, host=host, binary="nova-compute", forced=False
                )
            except openstack.exceptions.BadRequestException:
                logger.error(
                    f"Unable to force up host {host} as `done` evacuation migration "
                    "records remain associated with the host. Ensure the compute service "
                    "has been restarted, allowing these records to move to `completed` "
                    "before retrying this request."
                )
                return

        logger.info(f"Enabling nova-compute binary @ {host} ({service.id})")
        conn.compute.enable_service(
            service=service.id,
            host=host,
            binary="nova-compute",
        )


class ComputeDisable(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeDisable, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host to be disabled",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        conn = get_cloud_connection()

        services = conn.compute.services(**{"host": host, "binary": "nova-compute"})
        service = next(services)
        logger.info(f"Disabling nova-compute binary @ {host} ({service.id})")
        conn.compute.disable_service(
            service=service.id,
            host=host,
            binary="nova-compute",
            disabled_reason="MAINTENANCE",
        )


class ComputeList(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeList, self).get_parser(prog_name)
        parser.add_argument(
            "--project",
            default=None,
            type=str,
            help="Filter by project ID",
        )
        parser.add_argument(
            "--domain",
            default=None,
            type=str,
            help="Filter by domain ID",
        )
        parser.add_argument(
            "--details",
            default=False,
            help="Show details",
            action="store_true",
        )
        parser.add_argument(
            "host",
            nargs="?",
            type=str,
            help="Host on that all running instances are to be listed",
            default=None,
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host
        conn = get_cloud_connection()
        domain = parsed_args.domain
        project = parsed_args.project
        details = parsed_args.details

        result = []
        if host:
            for server in conn.compute.servers(all_projects=True, node=host):
                if project and server.project_id == project:
                    result.append([server.id, server.name, server.status])
                elif domain:
                    server_project = get_cloud_project(server.project_id)
                    if server_project.domain_id == domain:
                        result.append([server.id, server.name, server.status])
                else:
                    result.append([server.id, server.name, server.status])

            print(
                tabulate(
                    result,
                    headers=["ID", "Name", "Status"],
                    tablefmt="psql",
                )
            )

        else:
            hypervisors = conn.compute.hypervisors(details=True)
            if details:
                for hypervisor in hypervisors:
                    if hypervisor.get("uptime"):
                        try:
                            uptime = parse("uptime", hypervisor.get("uptime"))
                        except:
                            uptime = None
                    else:
                        uptime = None

                    if not uptime:
                        uptime = {
                            "uptime": "-",
                            "load_1m": "-",
                            "load_5m": "-",
                            "load_15m": "-",
                        }

                    result.append(
                        [
                            hypervisor.get("id"),
                            hypervisor.name,
                            hypervisor.get("status"),
                            hypervisor.get("state"),
                            uptime["uptime"],
                            uptime["load_1m"],
                            uptime["load_5m"],
                            uptime["load_15m"],
                        ]
                    )

                print(
                    tabulate(
                        result,
                        headers=[
                            "ID",
                            "Host",
                            "Status",
                            "State",
                            "Uptime",
                            "Load 1",
                            "Load 5",
                            "Load 15",
                        ],
                        tablefmt="psql",
                    )
                )
            else:
                for hypervisor in hypervisors:
                    result.append(
                        [
                            hypervisor.get("id"),
                            hypervisor.name,
                            hypervisor.get("status"),
                            hypervisor.get("state"),
                        ]
                    )

                print(
                    tabulate(
                        result,
                        headers=["ID", "Host", "Status", "State"],
                        tablefmt="psql",
                    )
                )


class ComputeEvacuate(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeEvacuate, self).get_parser(prog_name)
        parser.add_argument(
            "--yes",
            default=False,
            help="Always say yes",
            action="store_true",
        )
        parser.add_argument(
            "--target",
            default=None,
            type=str,
            help="Host to which all running instances are to be migrated",
        )
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host on that all running instances are to be migrated",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        target = parsed_args.target
        yes = parsed_args.yes
        conn = get_cloud_connection()

        result = []
        for server in conn.compute.servers(all_projects=True, node=host):
            result.append([server.id, server.name, server.status])

        if yes:
            answer = "yes"
        else:
            answer = prompt(f"Evacuate all servers on host {host} [yes/no]: ")

        start = []
        if answer in ["yes", "y"]:
            for server in result:
                if server[2] not in ["ACTIVE", "SHUTOFF"]:
                    logger.info(
                        f"{server[0]} ({server[1]}) in status {server[2]} cannot be evacuated"
                    )
                    continue
                if server[2] in ["ACTIVE"]:
                    logger.info(f"Stopping server {server[0]}")
                    start.append(str(server[0]))
                    conn.compute.stop_server(server[0])
                    inner_wait = True
                    while inner_wait:
                        time.sleep(2)
                        s = conn.compute.get_server(server[0])
                        if s.status not in ["SHUTOFF"]:
                            logger.info(
                                f"Stopping of {server[0]} ({server[1]}) is still in progress"
                            )
                            inner_wait = True
                        else:
                            inner_wait = False

            services = conn.compute.services(**{"host": host, "binary": "nova-compute"})
            service = next(services)
            logger.info(f"Forcing down nova-compute binary @ {host} ({service.id})")
            conn.compute.update_service_forced_down(
                service=service.id, host=host, binary="nova-compute", forced=True
            )

            for server in result:
                if server[2] in ["ACTIVE", "SHUTOFF"]:
                    logger.info(f"Evacuating server {server[0]}")
                    conn.compute.evacuate_server(server[0], host=target)

            if result:
                logger.info("Waiting 30 seconds")
                time.sleep(30)

            for server in start:
                logger.info(f"Starting server {server}")
                conn.compute.start_server(server)
                inner_wait = True
                while inner_wait:
                    time.sleep(2)
                    s = conn.compute.get_server(server)
                    if s.status not in ["ACTIVE"]:
                        logger.info(
                            f"Starting of {s.id} ({s.name}) is still in progress"
                        )
                        inner_wait = True
                    else:
                        inner_wait = False

            logger.info(f"Disabling nova-compute binary @ {host} ({service.id})")
            conn.compute.disable_service(
                service=service.id,
                host=host,
                binary="nova-compute",
                disabled_reason="EVACUATE",
            )


class ComputeMigrate(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeMigrate, self).get_parser(prog_name)
        parser.add_argument(
            "--yes",
            default=False,
            help="Always say yes",
            action="store_true",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait for completion of migration (Resize of cold migrated instances will not be confirmed!)",
            action="store_true",
        )
        parser.add_argument(
            "--no-cold-migration",
            default=False,
            help="Do not cold migrate instances",
            action="store_true",
        )
        parser.add_argument(
            "--force",
            default=False,
            help="Force a live-migration by not verifying the provided destination host by the scheduler. ",
            action="store_true",
        )
        parser.add_argument(
            "--target",
            default=None,
            type=str,
            help="Host to which all running instances are to be migrated",
        )
        parser.add_argument(
            "--project",
            default=None,
            type=str,
            help="Filter by project ID",
        )
        parser.add_argument(
            "--domain",
            default=None,
            type=str,
            help="Filter by domain ID",
        )
        parser.add_argument(
            "--filter",
            default=None,
            type=str,
            help="Filter by string",
        )
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host on that all running instances are to be migrated",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        target = parsed_args.target
        force = parsed_args.force
        no_wait = parsed_args.no_wait
        no_cold_migration = parsed_args.no_cold_migration
        yes = parsed_args.yes
        domain = parsed_args.domain
        project = parsed_args.project
        xfilter = parsed_args.filter

        conn = get_cloud_connection()

        result = []
        for server in conn.compute.servers(all_projects=True, node=host):
            if project and server.project_id == project:
                result.append([server.id, server.name, server.status])
            elif domain:
                server_project = get_cloud_project(server.project_id)
                if server_project.domain_id == domain:
                    result.append([server.id, server.name, server.status])
            elif xfilter:
                if xfilter in server.name:
                    result.append([server.id, server.name, server.status])
            else:
                result.append([server.id, server.name, server.status])

        if not result:
            logger.info(f"No migratable instances found on node {host}")

        for server in result:
            if server[2] in ["ACTIVE", "PAUSED"]:
                migration_type = "live"
            elif server[2] in ["SHUTOFF"] and not no_cold_migration:
                migration_type = "cold"
            else:
                logger.info(
                    f"{server[0]} ({server[1]}) in status {server[2]} cannot be migrated"
                )
                continue

            if yes:
                answer = "yes"
            else:
                answer = prompt(
                    f"{migration_type.capitalize()} migrate server {server[0]} ({server[1]}) [yes/no]: "
                )

            if answer in ["yes", "y"]:
                logger.info(
                    f"{migration_type.capitalize()} migrating server {server[0]}"
                )
                if migration_type == "live":
                    conn.compute.live_migrate_server(
                        server[0], host=target, block_migration="auto", force=force
                    )
                elif migration_type == "cold":
                    conn.compute.migrate_server(server[0], host=target)

                if not no_wait:
                    while True:
                        time.sleep(2)
                        s = conn.compute.get_server(server[0])
                        if (
                            migration_type == "live"
                            and s.status in ["MIGRATING"]
                            or migration_type == "cold"
                            and s.status in ["RESIZE"]
                        ):
                            logger.info(
                                f"{migration_type.capitalize()} migration of {server[0]} ({server[1]}) is still in progress"
                            )
                        elif migration_type == "cold" and s.status in ["VERIFY_RESIZE"]:
                            try:
                                conn.compute.confirm_server_resize(s)
                                logger.info(
                                    f"{migration_type.capitalize()} migration of {server[0]} ({server[1]}) confirmed"
                                )
                            except Exception as exc:
                                logger.error(
                                    f"{migration_type.capitalize()} migration of {server[0]} ({server[1]}) could not be confirmed"
                                )
                                raise exc
                            # NOTE: There seems to be no simple way to check whether the resize
                            # has been confirmed. The state is still "VERIFY_RESIZE" afterwards.
                            # Therefore we drop out without waiting for the "SHUTOFF" state
                            break
                        else:
                            logger.info(
                                f"{migration_type.capitalize()} migration of {server[0]} ({server[1]}) completed with status {s.status}"
                            )
                            break


class ComputeMigrationList(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeMigrationList, self).get_parser(prog_name)
        parser.add_argument(
            "--host",
            default=None,
            type=str,
            help="Only list migrations with the given host as source or destination",
        )
        parser.add_argument(
            "--server",
            default=None,
            type=str,
            help="Only list migrations for the given instance (name or ID)",
        )
        parser.add_argument(
            "--user",
            default=None,
            type=str,
            help="Only list migrations for the given user (name or ID)",
        )
        parser.add_argument(
            "--user-domain",
            default=None,
            type=str,
            help="Domain the user belongs to (name or ID)",
        )
        parser.add_argument(
            "--project",
            default=None,
            type=str,
            help="Only list migrations for the given project (name or ID)",
        )
        parser.add_argument(
            "--project-domain",
            default=None,
            type=str,
            help="Domain the project belongs to (name or ID)",
        )
        parser.add_argument(
            "--status",
            default=None,
            type=str,
            help="Only list migrations with the given status",
        )
        parser.add_argument(
            "--type",
            default=None,
            choices=["migration", "live-migration", "evacuation", "resize"],
            type=str,
            help="Only list migrations with the given type",
        )
        parser.add_argument(
            "--changes-since",
            default=None,
            type=datetime.datetime.fromisoformat,
            help="Only list migrations last chganged since the given date in ISO 8601 format (CCYY-MM-DDThh:mm:ss±hh:mm)",
        )
        parser.add_argument(
            "--changes-before",
            default=None,
            type=datetime.datetime.fromisoformat,
            help="Only list migrations last chganged before the given date in ISO 8601 format (CCYY-MM-DDThh:mm:ss±hh:mm)",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host
        server = parsed_args.server
        user = parsed_args.user
        user_domain = parsed_args.user_domain
        project = parsed_args.project
        project_domain = parsed_args.project_domain
        status = parsed_args.status
        migration_type = parsed_args.type
        changes_since = parsed_args.changes_since
        changes_before = parsed_args.changes_before

        if changes_before and changes_since:
            if not changes_since <= changes_before:
                logger.error(
                    "changes-since needs to be less or equal to changes-before"
                )
                return

        conn = get_cloud_connection()

        user_id = None
        if user:
            user_query = {}

            if user_domain:
                u_d = conn.identity.find_domain(user_domain, ignore_missing=True)
                if u_d and "id" in u_d:
                    user_query = dict(domain_id=u_d.id)
                else:
                    logger.error(f"No domain found for {user_domain}")
                    return

            u = conn.identity.find_user(user, ignore_missing=True, **user_query)
            if u and "id" in u:
                user_id = u.id
            else:
                logger.error(f"No user found for {user}")
                return

        project_id = None
        if project:
            project_query = {}

            if project_domain:
                p_d = conn.identity.find_domain(project_domain, ignore_missing=True)
                if p_d and "id" in p_d:
                    project_query = dict(domain_id=p_d.id)
                else:
                    logger.error(f"No domain found for {project_domain}")
                    return

            p = conn.identity.find_project(
                project, ignore_missing=True, **project_query
            )
            if p and "id" in p:
                project_id = p.id
            else:
                logger.error(f"No project found for {project}")
                return

        instance_uuid = None
        if server:
            try:
                s = conn.compute.find_server(
                    server, details=False, ignore_missing=False, all_projects=True
                )
                if s and "id" in s:
                    instance_uuid = s.id
                else:
                    raise openstack.exceptions.NotFoundException
            except openstack.exceptions.DuplicateResource:
                logger.error(f"Multiple servers where found for {server}")
                return
            except openstack.exceptions.NotFoundException:
                logger.error(f"No server found for {server}")
                return

        query = {}
        if host:
            query.update(dict(host=host))
        if instance_uuid:
            query.update(dict(instance_uuid=instance_uuid))
        if status:
            query.update(dict(status=status))
        if migration_type:
            query.update(dict(migration_type=migration_type))
        if user_id:
            query.update(dict(user_id=user_id))
        if project_id:
            query.update(dict(project_id=project_id))
        if changes_since:
            query.update(dict(changes_since=changes_since))
        if changes_before:
            query.update(dict(changes_before=changes_before))

        migrations = conn.compute.migrations(**query)
        result = [
            [
                m.source_compute,
                m.dest_compute,
                m.status,
                m.migration_type,
                m["instance_uuid"],
                m.user_id,
                m.created_at,
                m.updated_at,
            ]
            for m in migrations
        ]

        print(
            tabulate(
                result,
                headers=[
                    "Source",
                    "Destintion",
                    "Status",
                    "Type",
                    "Server UUID",
                    "User",
                    "Created At",
                    "Updated At",
                ],
                tablefmt="psql",
            )
        )


class ComputeStart(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeStart, self).get_parser(prog_name)
        parser.add_argument(
            "--yes",
            default=False,
            help="Always say yes",
            action="store_true",
        )
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host on that all running instances are to be started",
        )
        return parser

    def take_action(self, parsed_args):
        yes = parsed_args.yes
        host = parsed_args.host[0]
        conn = get_cloud_connection()

        result = []
        for server in conn.compute.servers(all_projects=True, node=host):
            result.append([server.id, server.name, server.status])

        for server in result:
            if server[2] not in ["SHUTOFF"]:
                logger.info(
                    f"{server[0]} ({server[1]}) in status {server[2]} cannot be started"
                )
                continue

            if yes:
                answer = "yes"
            else:
                answer = prompt(f"Start server {server[0]} ({server[1]}) [yes/no]: ")

            if answer in ["yes", "y"]:
                logger.info(f"Starting server {server[0]}")
                conn.compute.start_server(server[0])


class ComputeStop(Command):
    def get_parser(self, prog_name):
        parser = super(ComputeStop, self).get_parser(prog_name)
        parser.add_argument(
            "--yes",
            default=False,
            help="Always say yes",
            action="store_true",
        )
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Host on that all running instances are to be stopped",
        )
        return parser

    def take_action(self, parsed_args):
        yes = parsed_args.yes
        host = parsed_args.host[0]
        conn = get_cloud_connection()

        result = []
        for server in conn.compute.servers(all_projects=True, node=host):
            result.append([server.id, server.name, server.status])

        for server in result:
            if server[2] not in ["ACTIVE", "PAUSED"]:
                logger.info(
                    f"{server[0]} ({server[1]}) in status {server[2]} cannot be stopped"
                )
                continue

            if yes:
                answer = "yes"
            else:
                answer = prompt(f"Stop server {server[0]} ({server[1]}) [yes/no]: ")

            if answer in ["yes", "y"]:
                logger.info(f"Stopping server {server[0]}")
                conn.compute.stop_server(server[0])



# SPDX-License-Identifier: Apache-2.0

import os
import subprocess

from cliff.command import Command
from loguru import logger
import yaml

from osism.tasks import conductor, netbox, handle_task
from osism import utils


class Ironic(Command):
    def get_parser(self, prog_name):
        parser = super(Ironic, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        parser.add_argument(
            "--task-timeout",
            default=os.environ.get("OSISM_TASK_TIMEOUT", 300),
            type=int,
            help="Timeout for a scheduled task that has not been executed yet",
        )
        parser.add_argument(
            "--force-update",
            help="Force update of baremetal nodes (Used to update non-comparable items like passwords)",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait
        task_timeout = parsed_args.task_timeout

        task = conductor.sync_ironic.delay(force_update=parsed_args.force_update)
        if wait:
            logger.info(
                f"Task {task.task_id} (sync ironic) is running in background. Output comming soon."
            )
            try:
                return utils.fetch_task_output(task.id, timeout=task_timeout)
            except TimeoutError:
                logger.error(
                    f"Timeout while waiting for further output of task {task.task_id} (sync ironic)"
                )
        else:
            logger.info(
                f"Task {task.task_id} (sync ironic) is running in background. No more output."
            )


class Sync(Command):
    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = conductor.sync_netbox.delay()
        if wait:
            logger.info(
                f"Task {task.task_id} (sync netbox) is running. Wait. No more output."
            )
            task.wait(timeout=None, interval=0.5)


class Manage(Command):
    def get_parser(self, prog_name):
        parser = super(Manage, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the management of the NetBox has been completed",
            action="store_true",
        )
        parser.add_argument(
            "--no-netbox-wait",
            default=False,
            help="Do not wait for the NetBox API to be ready",
            action="store_true",
        )
        parser.add_argument(
            "--parallel",
            type=str,
            default=None,
            help="Process up to n files in parallel",
        )
        parser.add_argument(
            "--limit",
            type=str,
            default=None,
            help="Limit files by prefix",
        )
        parser.add_argument(
            "--skipdtl",
            default=False,
            help="Skip devicetype library",
            action="store_true",
        )
        parser.add_argument(
            "--skipmtl",
            default=False,
            help="Skip moduletype library",
            action="store_true",
        )
        parser.add_argument(
            "--skipres",
            default=False,
            help="Skip resources",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait
        arguments = ["run"]

        if parsed_args.no_netbox_wait:
            arguments.append("--no-wait")
        else:
            arguments.append("--wait")

        if parsed_args.parallel:
            arguments.append("--parallel")
            arguments.append(parsed_args.parallel)

        if parsed_args.limit:
            arguments.append("--limit")
            arguments.append(parsed_args.limit)

        if parsed_args.skipdtl:
            arguments.append("--skipdtl")
        else:
            arguments.append("--no-skipdtl")

        if parsed_args.skipmtl:
            arguments.append("--skipmtl")
        else:
            arguments.append("--no-skipmtl")

        if parsed_args.skipres:
            arguments.append("--skipres")
        else:
            arguments.append("--no-skipres")

        task_signature = netbox.manage.si(*arguments)
        task = task_signature.apply_async()
        if wait:
            logger.info(
                f"It takes a moment until task {task.task_id} (netbox-manager) has been started and output is visible here."
            )

        return handle_task(task, wait, format="script", timeout=3600)


class Versions(Command):
    def get_parser(self, prog_name):
        parser = super(Versions, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        task = netbox.ping.delay()
        task.wait(timeout=None, interval=0.5)
        result = task.get()
        print(result)


class Console(Command):
    def get_parser(self, prog_name):
        parser = super(Console, self).get_parser(prog_name)
        parser.add_argument(
            "type",
            nargs=1,
            choices=["info", "search", "filter", "shell"],
            help="Type of the console (default: %(default)s)",
        )
        parser.add_argument(
            "arguments", nargs="*", type=str, default="", help="Additional arguments"
        )

        return parser

    def take_action(self, parsed_args):
        type_console = parsed_args.type[0]
        arguments = " ".join(
            [f"'{item}'" if " " in item else item for item in parsed_args.arguments]
        )

        home_dir = os.path.expanduser("~")
        nbcli_dir = os.path.join(home_dir, ".nbcli")
        if not os.path.exists(nbcli_dir):
            os.mkdir(nbcli_dir)

        nbcli_file = os.path.join(nbcli_dir, "user_config.yml")
        if not os.path.exists(nbcli_file):
            try:
                with open("/run/secrets/NETBOX_TOKEN", "r") as fp:
                    token = fp.read().strip()
            except FileNotFoundError:
                token = None

            url = os.environ.get("NETBOX_API", None)

            if not token or not url:
                logger.error("NetBox integration not configured.")
                return

            subprocess.call(
                ["/usr/local/bin/nbcli", "init"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            os.remove(nbcli_file)

            nbcli_config = {
                "pynetbox": {
                    "url": url,
                    "token": token,
                },
                "requests": {"verify": False},
                "nbcli": {"filter_limit": 50},
                "user": {},
            }
            with open(nbcli_file, "w") as fp:
                yaml.dump(nbcli_config, fp, default_flow_style=False)

        subprocess.call(f"/usr/local/bin/nbcli {type_console} {arguments}", shell=True)
